;; Generated package description from tmr.el  -*- no-byte-compile: t -*-
(define-package "tmr" "0.2.3.0.20220517.83449" "Set timers using a convenient notation" '((emacs "27.1")) :commit "ca6c0c33db91125e1fe5edf4221dd9b9e3ecfe3e" :authors '(("Protesilaos Stavrou" . "info@protesilaos.com") ("Damien Cassou" . "damien@cassou.me")) :maintainer '("Protesilaos Stavrou" . "info@protesilaos.com") :keywords '("convenience" "timer") :url "https://git.sr.ht/~protesilaos/tmr")
